from json.decoder import JSONDecodeError
from logging import INFO, DEBUG, WARNING, ERROR, CRITICAL, NOTSET, root, getLogger, NullHandler, Formatter, \
    StreamHandler, FileHandler, Logger
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from copy import deepcopy
from os import mkdir, getenv
from tools import Singleton
from os.path import join as join_path, exists as exists_path
from typing import List
# from .handlers import DBHandler
import pymysql
from jsonschema import validate as schema_validate, ValidationError
from json import loads as json_loads

root.setLevel(ERROR)
root.addHandler(StreamHandler())

_TIMED_ROTATION = 1
_SIZE_ROTATION = 2

_nameToLevel = {
    'CRITICAL': CRITICAL,
    'ERROR': ERROR,
    'WARN': WARNING,
    'WARNING': WARNING,
    'INFO': INFO,
    'DEBUG': DEBUG,
    'NOTSET': NOTSET,
}
_rotationFileTypes = {
    "TIMED": _TIMED_ROTATION,
    "SIZE": _SIZE_ROTATION
}


class LoggingConfig:
    def __init__(self, config, config_connection_db):
        self.__config = config
        if config_connection_db:
            self.__user_connection = config_connection_db['user']
            self.__password_connection = config_connection_db['password']
            self.__host_connection = config_connection_db['host']
            self.__port_connection = config_connection_db['port']
            self.__db_name_connection = config_connection_db['DB_name']
        self.__generated_loggers: List[str] = []

        self.__general_logger = None
        self.__logger = self.get_logger('LoggingConfig')
        self.__logger.debug(f'Значения параметров:\n'
                            f'config: {config}\n'
                            f'config_connection_db: {config_connection_db}', stack_info=True)
        self.__general_logger = self.get_logger('General')

    def __parse_file_handler__(self, config):
        if 'logging_rotation_type' not in config:
            config['logging_rotation_type'] = None

        if not exists_path(config['directory_log']):
            mkdir(config['directory_log'])

        if _rotationFileTypes[config['logging_rotation_type']] == _TIMED_ROTATION:
            when, interval = config['logging_interval'].split(' ')
            handler = TimedRotatingFileHandler(config['file_name'], when=when, backupCount=config['backup_count'],
                                               interval=int(interval))
        elif _rotationFileTypes[config['logging_rotation_type']] == _SIZE_ROTATION:
            handler = RotatingFileHandler(config['file_name'], maxBytes=int(config['logging_interval']),
                                          backupCount=config['backup_count'])
        else:
            handler = FileHandler(config['file_name'])

        handler.setFormatter(Formatter(config['format_logging']))
        handler.setLevel(config['level_logging'])

        return handler

    def __parse_db_handler(self, config):
        connection = pymysql.Connection(host=self.__host_connection, user=self.__user_connection,
                                        password=self.__password_connection, port=self.__port_connection,
                                        db=self.__db_name_connection, cursorclass=pymysql.cursors.DictCursor)
        # handler = DBHandler(connection, 'logs')
        #
        # handler.setFormatter(Formatter(config['format_logging']))
        # handler.setLevel(config['level_logging'])
        #
        # return handler

    def __parse_console_handler(self, config):
        handler = StreamHandler()
        handler.setLevel(config['level_logging'])
        handler.setFormatter(Formatter(config['format_logging']))

        return handler

    def __generation_handlers__(self, config_handlers):
        handlers = []
        for i in config_handlers:
            if 'enabled_handler' in i and i['enabled_handler'] is False:
                continue

            if i['type_handler'] == "file":
                handlers.append(self.__parse_file_handler__(i))
            elif i['type_handler'] == 'db':
                handlers.append(self.__parse_db_handler(i))
            elif i['type_handler'] == 'console':
                handlers.append(self.__parse_console_handler(i))

        if not handlers:
            handlers.append(NullHandler())

        return handlers

    def get_logger(self, logger_name) -> Logger:
        logger = getLogger(logger_name)

        if logger_name not in self.__generated_loggers:
            parent = None

            config_for_logger_name = None
            for i in self.__config['settings_overload']:
                if i['name_logger'] == logger_name:
                    config_for_logger_name = i['settings']
                    parent = i.get('parent')
                    check_keys = ['level_logging', 'format_logging', 'handlers']
                    for key in check_keys:
                        if key not in config_for_logger_name.keys():
                            config_for_logger_name[key] = self.__config[key]
                    break

            if config_for_logger_name is None:
                config_for_logger_name = deepcopy(self.__config)
                del config_for_logger_name['settings_overload']
            for i in config_for_logger_name['handlers']:
                if 'directory_log' in i:
                    i['file_name'] = join_path(i['directory_log'], f"{logger_name}.log")

                check_keys = ['level_logging', 'format_logging']
                for key in check_keys:
                    if key not in i:
                        i[key] = config_for_logger_name[key]

            logger.setLevel(config_for_logger_name['level_logging'])

            handlers = self.__generation_handlers__(config_for_logger_name['handlers'])

            for i in handlers:
                logger.addHandler(i)

            if parent:
                logger.parent = self.get_logger(parent)
            elif self.__general_logger is not None:
                logger.parent = self.__general_logger

            self.__generated_loggers.append(logger_name)
        return logger


class Configuration(metaclass=Singleton):
    """
    Класс конфигов приложения
    """

    __FILE_TEMPLATE_CONFIG = getenv("TEMPLATE_CONFIG")
    __FILE_CONFIG = getenv("CONFIG_FILE")

    def __init__(self):
        try:
            self.__logger = root
            if not Configuration.__FILE_CONFIG:
                raise KeyError("Не задана переменная файла конфигурации \"CONFIG_FILE\"")
            if not Configuration.__FILE_TEMPLATE_CONFIG:
                raise KeyError("Не задана переменная файла шаблона конфигурации \"TEMPLATE_CONFIG\"")

            self.__path_to_config_file = Configuration.__FILE_CONFIG
            with open(self.__path_to_config_file, 'r') as file_configuration:
                configuration = json_loads(file_configuration.read())

            self.__check_struct_config__()
            for i in configuration:
                self.__logger.debug(f"Конфигурация {i}: {configuration[i]}")
                self.__setattr__(i, configuration[i])
            self.logging = LoggingConfig(self.logging, self.DB)
            self.__logger = self.logging.get_logger("Configuration")
            self.__logger.debug(f"Путь до конфигурации: {self.path_to_config_file}")
            self.__logger.debug(f'Путь до шаблона конфигурации: {Configuration.__FILE_TEMPLATE_CONFIG}')
        except (JSONDecodeError, KeyError, ValidationError) as e:
            self.__logger.error(str(e), exc_info=True)
            exit(1)

    def __getattr__(self, item):
        return self.__dict__.get(item)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)

    def __check_struct_config__(self):
        self.__logger.info(f'Проверка конфигурации по шаблону')
        with open(Configuration.__FILE_CONFIG, 'r') as config_file:
            with open(Configuration.__FILE_TEMPLATE_CONFIG, 'r') as template_config:
                schema_validate(json_loads(config_file.read()), json_loads(template_config.read()))
                self.__logger.info(f'Проверка конфигурации пройдена')

    @property
    def path_to_config_file(self):
        self.__logger.debug(f'Возвращение значения: {self.__path_to_config_file}', stack_info=True)
        return self.__path_to_config_file



"""
Модуль с описанием класса Configuration
"""
# from .handlers import DBHandler
from .configuration import Configuration

__version__ = '2.3.2'
__name__ = 'py-configuration-Orinnass'

# TODO: изменить минимальный процент покрытия кода тестами



"""
Модуль для работы с конфигурацией из командной строки
"""
from sys import argv
from os import getenv
from os.path import exists as path_exists, isfile
from json import dumps as json_dump, loads as json_load
import argparse
from config import __version__

new_element = None

# TODO: сделать обработку ref


def __dump_json__(file_path, json):
    with open(file_path, 'w') as file:
        file.write(json_dump(json, indent=4))


def __parse_element__(element, name_parent_element, parent_element):
    global new_element
    if element['type'] == 'object':
        parent_element[name_parent_element] = {}
        parent_element = parent_element[name_parent_element]
        for i in element['required']:
            __parse_element__(element['properties'][i], i, parent_element)
    else:
        parent_element[name_parent_element] = element.get('default')


def __update_add_command__(params, template):
    # TODO: сделать проверку на существование элемента и запрашивать перезапись
    with open(params.config_file, 'r') as config_file:
        config = json_load(config_file.read())
    for i in params.elements:
        if i not in template['properties']:
            print(f'Элемента {i} нет в шаблоне')
            return
        global new_element
        new_element = {}
        __parse_element__(template['properties'][i], i, new_element)
        config[i] = new_element[i]

    __dump_json__(params.config_file, config)
    print(f"{'Элемент добавлен' if len(params.elements) == 1 else 'Элементы добавлены'}")


def __update_delete_command__(params, template):
    with open(params.config_file, 'r') as config_file:
        config = json_load(config_file.read())

    for i in params.elements:
        if i in config.keys():
            del config[i]
            print(f"{i} элемент удален")
        else:
            print(f"{i} элемент отсутсвует в конфиге")

    __dump_json__(params.config_file, config)

    print('Удаление выполнено')


def create_command(params):
    """
    Команда создания конфига из шаблона
    """
    default_config = {
        "logging": {
            "level_logging": "ERROR",
            "format_logging": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            "handlers": [
                {
                    "type_handler": "file",
                    "level_logging": "ERROR",
                    "directory_log": "logs",
                    "logging_rotation_type": "TIMED",
                    "backup_count": 1,
                    "logging_interval": "d 1"
                },
                {
                    "enabled_handler": False,
                    "type_handler": "db",
                    "level_logging": "ERROR"
                },
                {
                    "type_handler": "console",
                    "level_logging": "ERROR"
                }
            ],
            "settings_overload": [
                {
                    "name_logger": "General",
                    "settings": {
                        "handlers": [
                            {
                                "enabled_handler": True,
                                "type_handler": "file",
                                "directory_log": "logs",
                                "logging_rotation_type": "TIMED",
                                "backup_count": 1,
                                "logging_interval": "d 1"
                            }
                        ]
                    }
                }
            ]
        }
    }

    if params.list_elements:
        if not getenv('TEMPLATE_CONFIG', default=None):
            print("Не определена переменная окружения TEMPLATE_CONFIG")
            return

        template = json_load(open(getenv("TEMPLATE_CONFIG")).read())

        message = "Список доступных элементов: "
        for i in template['properties']:
            message += f"{i}, "
        print(message[:-2])
        return
    if path_exists(params.config_file):
        print('Файл конфигурации уже существует, перезаписать его? (y/n)')
        answer = input()
        if answer.lower() != "yes" and answer.lower() != 'y':
            return
    if params.elements:
        if not getenv('TEMPLATE_CONFIG', default=None):
            print("Не определена переменная окружения TEMPLATE_CONFIG")
            return

        template = json_load(open(getenv("TEMPLATE_CONFIG")).read())

        for i in params.elements:
            if i not in template['properties']:
                print(f'Элемента {i} нет в шаблоне')
                return
            global new_element
            new_element = {}
            __parse_element__(template['properties'][i], i, new_element)
            default_config[i] = new_element[i]

    __dump_json__(params.config_file, default_config)
    print('Конфигурация создана')


def update_command(params):
    """
    Команда обновления конфига
    """
    if not getenv('TEMPLATE_CONFIG', default=None):
        print("Не определена переменная окружения TEMPLATE_CONFIG")
        return
    if not path_exists(params.config_file) and not isfile(params.config_file):
        print("Указанного файла не существует")
        return

    # TODO: в удаление шаблон не используется исправить
    template = json_load(open(getenv("TEMPLATE_CONFIG")).read())
    update_methods = {
        "add": __update_add_command__,
        "delete": __update_delete_command__
    }
    update_methods[params.update_command](params, template)


methods = {
    "create": create_command,
    "update": update_command
}


def create_parse():
    """
    Команда создания парсинга аргументов
    """
    parser = argparse.ArgumentParser(add_help=False)

    parent_group = parser.add_argument_group(title="Параметры")
    parent_group.add_argument('--version', action='version', help='Вывести номер версии',
                              version='%(prog)s {}'.format(__version__))
    parent_group.add_argument("--help", "-h", action="help", help="Справка")

    subparsers = parser.add_subparsers(dest="command", title="Возможные команды",
                                       description="Команды, которые должны быть в качестве первого параметра %(prog)s")

    create_command_parser = subparsers.add_parser("create", add_help=False)
    create_command_parser.add_argument('config_file')
    create_command_parser.add_argument('-e', '--elements', nargs='+', help="Элементы, которые нужно добавить в конфиг. "
                                                                           "Для работы этого параметра требуется "
                                                                           "указать переменную окружения "
                                                                           "TEMPLATE_CONFIG")
    create_command_parser.add_argument('--list-elements', action='store_true', default=False)
    create_command_parser.add_argument('-h', '--help', action='help', help='Справка')

    update_command_parser = subparsers.add_parser('update', add_help=False)
    update_command_parser.add_argument('config_file')
    update_command_parser.add_argument('update_command', choices=['add', 'delete'])
    update_command_parser.add_argument('elements', nargs='+')
    update_command_parser.add_argument('-h', '--help', action='help', help='Справка')

    return parser


if __name__ == '__main__':
    main_parser = create_parse()
    parsed_params = main_parser.parse_args(argv[1:])

    methods[parsed_params.command](parsed_params)



from config.configuration import Configuration
from flask import Flask, Response
from json import dumps

__app = Flask(__name__)


@__app.route('/api/config/get')
def __get_config__():
    config = Configuration()
    response = dumps(config)
    del config
    return Response(response, mimetype='application/json')


def start_server():
    config = Configuration()

    if config.config_server_api['enabled']:
        host = config.config_server_api['address']
        port = config.config_server_api['port']
        del config
        __app.run(host, port=port)


if __name__ == '__main__':
    start_server()



# from pymysql import Connection as pymySQLConnection
# from logging import Handler, LogRecord
#
#
# class DBHandler(Handler):
#     """
#     Обработчик логирования в БД
#     """
#     # TODO: Дописать обработчик
#     def __init__(self, connection: pymySQLConnection, table: str):
#         super().__init__()
#         self.__connection = connection
#         self.__cursor = connection.cursor()
#         self.__table = table
#
#     def emit(self, record: LogRecord) -> None:
#         self.__cursor.execute(f"insert into `{self.__table}` (`logger`, `level_log`, `message`, `time`) "
#                               f"values (%s, %s, %s, %s)", [record.name, record.levelname, record.message, record.asctime])
#         self.__connection.commit()
